/**
 * This class extends JPanel.<br/>
 * It is used as the renderer of the images associated to the<br/>
 * nodes of the tree.
 *
 *
 */

package Alma.animation;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.font.FontRenderContext;
import java.awt.font.TextLayout;
import java.awt.geom.AffineTransform;
import java.awt.geom.Line2D;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.JPanel;

/**
 *
 * @author Nuno Oliveira
 */
public class DrawingPanel extends JPanel {
    private AffineTransform at; //used to perform transformatio in the grphics
    private ArrayList<Actor> actors; //used to store the actors that will be painted in the panel
    private ArrayList<Pair> actorsPaint; //used to store the list of pairs (Actor, PoseNumber)
                                         //to be painted in the panel.

    public DrawingPanel() {
        this.setBackground(Color.WHITE);
        this.at = new AffineTransform();
        this.actors = new ArrayList<Actor>();
        this.actorsPaint = new ArrayList<Pair>();
        makeAnimation();
    }
    

    public DrawingPanel(ArrayList<Actor> objs) {
        this.setBackground(Color.WHITE);
        this.at = new AffineTransform();
        this.actors = objs;
        this.actorsPaint = new ArrayList<Pair>();
        makeAnimation();
    }

    public ArrayList<Actor> getActors() {
        return actors;
    }

    public void setActors(ArrayList<Actor> actors) {
        this.actors = actors;
    }

    

    /**
     * sets the renderer to the center of the panel
     */
    public void centerRenderer(){
        int w = this.getWidth();
        int h=this.getHeight();
        //the transformation is the identity:
        //so everything is set the default values;
        this.at.setToIdentity();
        //This translation transformation is used to set the
        //drawing render to the center of the Panel
        this.at.translate(w/2, h/2);
    }


    /**
     * Centers an image to the center of the drawing panel.<br>
     * It must be used before drawing the image;<br>
     * <br>
     * The final output, X, must be used with graphics.setTransform(X).<br>
     * to reset the transformation.
     *
     * @param i Image to center
     * @param g The graphics to draw the image in
     * @return the transformation in the graphics, before performing the centering;
     */
    public AffineTransform centerImage(BufferedImage i, Graphics2D g){
        //store the old (actual) transformation in the Graphics
        AffineTransform oldTransformation = g.getTransform();
        //center the grphic renderer
        centerRenderer();
        int w = i.getWidth();
        int h = i.getHeight();
        AffineTransform toCenter = new AffineTransform();
        //concatenate the transforations
        toCenter.concatenate(at);
        //translate the image up and left to center the image
        toCenter.translate(-(w/2), -(h/2));
        //apply the coordinates transformation to the graphics
        g.transform(toCenter);
        
        return oldTransformation;
    }

    /**
     * Method to draw the axes in the Panel that shows the animation
     * @param g The graphics used to draw the axes.
     */
    public void drawAxes(Graphics2D g){
        int w = this.getBounds().width;
        int h = this.getBounds().height;
        Line2D l;
        g.setPaint(Color.gray);
        g.setStroke(new BasicStroke(2));
        //Draw horizontal line
        l = new Line2D.Float(0, h/2, w, h/2);
        g.draw(l);

        //Draw vertical lines
        l = new Line2D.Float(w/2, 0, w/2, h);
        g.draw(l);
        g.setPaint(Color.BLACK);
        g.setStroke(new BasicStroke(1));
    }


    /**
     * Method to draw the axes in the Panel that shows the animation
     * @param g The graphics used to draw the axes.
     */
    public void drawGrid(Graphics2D g){
        int w = this.getBounds().width;
        int h = this.getBounds().height;
        Line2D l;
        g.setPaint(new Color(242, 242, 242));
        //Draw vertical lines after the center
        for(int i = w/2; i < w; i+=20){
            l = new Line2D.Float(i, 0, i, h);
            g.draw(l);
        }
        //Draw vertical lines before the center
        for(int i = w/2; i >0; i-=20){
            l = new Line2D.Float(i, 0, i, h);
            g.draw(l);
        }
        

        //Draw horizonatl lines before the center
        for(int i = h/2; i > 0; i-=20){
            l = new Line2D.Float(0, i, w, i);
            g.draw(l);
        }
        //Draw horizonatl lines after the center
        for(int i = h/2; i < h; i+=20){
            l = new Line2D.Float(0, i, w, i);
            g.draw(l);
        }
        g.setPaint(Color.BLACK);

    }


    /**
     * This method animates frame by frame (sleeping some ms between each frame)<br/>
     * The actors are processed by the order of their appearence<br/>
     * Only the last image in the sequence animation is reutilized when repainting</br>
     * <br/>
     * The algorithm is simple:<br/>
     * For each Actor in the list of actors of the DrawingPanel, we will loop<br/>
     * over their used poses (the poses that are really used by the actor)<br/>
     * Then, for each pair in the list of actors to paint, we will see if this actor<br/>
     * already exixts in the list. If it not exists, we insert the new pair <br/>
     * (Actor, poseNumber) in this list. We break the iteration otherwise and make 2 tests:<br/>
     * If the size of this list is 0, then we insert the actor.<br/>
     * If the actor existed in the list of actor to paint, we remove the pair, and set the new one<br/>
     * Then we repaint the panel.
     *
     */
    public void makeAnimation() {
        boolean foundActor;
        ArrayList<Actor> single = new ArrayList<Actor>();
        ArrayList<Actor> aggregated = new ArrayList<Actor>();
        this.actorsPaint.clear();
        //System.out.println(this.actors.toString());
        for(Actor a : this.actors){
            if(a.isIsAggregated()){
                aggregated.add(a);
            }
            else{
                //System.out.println("SINGLE");
                single.add(a);
            }
            //System.out.println("END_ITER");
        }
        //System.out.println("END_CICLO");
        //System.out.println("ACTORs IN SINGLE: " + single.toString());

        for(Actor ac : single){
            //System.out.println("NEW ACTOR: " + ac.toString());

            int i = 0;
            int counter = 0;
            foundActor = false;
                for(int p: ac.getUsedPoses()){
                    counter++;
                    //System.out.println("ENTRAR");
                    for(int k = 0; k < this.actorsPaint.size(); k++){
                        //System.out.println("DENTRO k = " + k );
                        //System.out.println(this.actorsPaint.toString());
                        if(!foundActor){
                            Actor a = (Actor)this.actorsPaint.get(k).getFst();
                            if(a==ac){
                                //System.out.print(a.toString());
                                //System.out.println(" = " + ac.toString());
                                foundActor = true;
                                break;
                                //this.actorsPaint.add(i, new Pair(ac, p));
                            }
                            i++;
                        }
                    }
                    //System.out.println(this.actorsPaint);
                    if(!foundActor && this.actorsPaint.size()!=0){
                        this.actorsPaint.add(new Pair(ac, p));
                        //System.out.println(this.actorsPaint);
                    }
                    //System.out.println("SAIR");
                    if(this.actorsPaint.size()==0){
                        this.actorsPaint.add(new Pair(ac, p));
                        //System.out.println(this.actorsPaint);
                    }
                    if(foundActor) {
                        //System.out.println(ac.toString());
                        this.actorsPaint.remove(i);
                        this.actorsPaint.add(i, new Pair(ac, p));
                    }

                    //System.out.println("Actors - " + this.actorsPaint);
                    //System.out.println("O p que vou animar e: " + p);
                    //paints the component once again (The repaint method does not work here as we expect)...
                    try{
                        //System.out.println("PAINTED");
                        this.paintComponent(this.getGraphics());
                    }
                    catch(NullPointerException e){
                        this.repaint();
                    }

                    //This if is used (along with counter) to see if the pose being processed is the last or not
                    //in case of being the last, the panel does not sleep, otherwise sleeps.
                    //In other words, the repaint of the frames are delayed only if there are more than one pose for
                    //the same actor, And when there are more than one, it does not sleep after painting the last actor.
                    if(ac.getUsedPoses().length >counter){
                        //System.out.println("vou dormir");
                        try {
                            Thread.sleep(700);
                        } catch (InterruptedException ex) {
                            Logger.getLogger(DrawingPanel.class.getName()).log(Level.SEVERE, null, ex);
                        }
                    }
                }
        }

        for(Actor ac: aggregated){
            this.actorsPaint.add(new Pair(ac, ac.getUsedPoses()[0]));
            try{
                this.paintComponent(this.getGraphics());
            }
            catch(NullPointerException e){
                this.repaint();
            }
        }

    }

   

    @Override
    public void paintComponent(Graphics g){
        super.paintComponent(g);
        AffineTransform reset;
        Graphics2D g2 = (Graphics2D)g;
        BufferedImage i;
        int w;
        int h;

        //draw the axes in the panel
        drawGrid(g2);
        drawAxes(g2);



        //sets a font render context to use for the label rendering
        FontRenderContext frc = g2.getFontRenderContext();

        //paint each actor in the scene
        //System.out.println("Iam Pinting...");
        Actor actor;
        int p;
        for(Pair pair : this.actorsPaint){
            actor = (Actor)pair.getFst();
            p = (Integer)pair.getSnd();
            i = actor.getPoses().get(p);
            w = i.getWidth();
            h = i.getHeight();
            //center the image
            reset = centerImage(i, g2);
            //translation should be (X,-Y) because Y-axe is inverted
            g2.translate(actor.getX()*20, -actor.getY()*20);

            //Put the label below the image (if the label exists)
            if(actor.isIsLabeled()){
                TextLayout tl2 = new TextLayout(actor.getLabel(), new Font("Helvetica", 3, 10), frc);
                float width2 = (float)tl2.getBounds().getWidth();
                //float height = (float)thisTl.getBounds().getHeight();
                tl2.draw(g2, w/2-width2/2, h + 10);
            }
            //Rotate after put the label, so the label appears alway below the image
            //the rotation angle should be inverted because Y-axe was inverted
            g2.rotate(Math.toRadians(-(actor.getAngle())),w/2,h/2);
            g2.drawImage(i, null, 0, 0);

            g2.setTransform(reset);
            //System.out.println("Pintei o p: " + p);

        }
                
    }


    @Override
    public Object clone(){
        DrawingPanel dp = new DrawingPanel();
        dp.setActors(this.getActors());
        return dp;
    }


}
